/**
  PasswordMaker - Creates and manages passwords
  Copyright (C) 2005 Eric H. Jung and LeahScape, Inc.
  http://passwordmaker.org/
  grimholtz@yahoo.com

  This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or (at
  your option) any later version.

  This library is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESSFOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
  for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
**/

// Helper functions and "classes" -- called by both passwordmaker.js and
// passwordmakerOverlay.js

/** Notice we only keep the encrypted version of the master password in memory.
 * This is a security measure so a remote debugger (if attached)
 * would have a harder time determining the master password. A remote
 * debugger (if connected) can still determine the master password even if it's encrypted,
 * but it's more difficult this way.
 *
 * The MasterPassword class...
 */
function MasterPassword(encryptedMasterPassword, key) {
  this.encryptedData = encryptedMasterPassword;
  this.key = key;
}

// Ctor
function PasswordMakerHelper() {
  this.initDataSource();

  // Get the accounts sequence, or create it if it doesn't exist
  this.accountsContainer = this.rdfContainerUtils.MakeSeq(this.datasource, passwordMaker.accountsSubjectRes);
  if (this.rdfContainerUtils.IsEmpty(this.datasource, passwordMaker.accountsSubjectRes)) {
    // Make the defaults resources using mozilla preferences we used in prior versions.
    // If they exist, this maintains the same defaults for users who are upgrading from
    // a prior release. If they don't exist, we just use hard-coded defaults.
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.whereLeetLBPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("whereLeet", "off")), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.leetLevelLBPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("leetLevel", 1)), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.hashAlgorithmLBPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("hashAlgorithm", "md5")), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.passwordLengthPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("passwordLength", "8")), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.namePredicateRes, this.rdfService.GetLiteral("Defaults"), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.descriptionPredicateRes, this.rdfService.GetLiteral("Default settings for URLs not elsewhere in this list"), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.protocolCBPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("protocol", "false")), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.subdomainCBPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("subdomain", "false")), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.domainCBPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("domain", "true")), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.pathCBPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("path", "false")), true);
    // This didn't exist before 0.6.1
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.charsetPredicateRes, this.rdfService.GetLiteral("0123456789abcdef"), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.prefixPredicateRes, this.rdfService.GetLiteral(""), true);
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.suffixPredicateRes, this.rdfService.GetLiteral(""), true);
    // 0.7.3--autopopulate became an account-level setting
    this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.autoPopulatePredicateRes, this.rdfService.GetLiteral("false"), true);
    // Stick the defaults resource in the accounts container
    this.accountsContainer.AppendElement(passwordMaker.defaultSubjectRes);
  }
  // Conversion to 0.7.3 from older versions where autoPop was a global setting
  if (!this.datasource.hasArcOut(passwordMaker.defaultSubjectRes, passwordMaker.autoPopulatePredicateRes)) {
    try {
      // same as passwordMaker._getLiteral
      var oldValue = this.datasource.GetTarget(passwordMaker.globalSettingsSubjectRes, passwordMaker.autoPopulatePredicateRes, true);
      if (oldValue != null) {
        oldValue = oldValue.QueryInterface(Components.interfaces.nsIRDFLiteral);
        oldValue = oldValue == null ? "false" : oldValue.Value;
      }
      if (oldValue == null)
        oldValue = "false";
      this.datasource.Assert(passwordMaker.defaultSubjectRes, passwordMaker.autoPopulatePredicateRes, this.rdfService.GetLiteral(oldValue), true);
      oldValue = this.datasource.GetTarget(passwordMaker.globalSettingsSubjectRes, passwordMaker.autoPopulatePredicateRes, true);
      this.datasource.Unassert(passwordMaker.globalSettingsSubjectRes, passwordMaker.autoPopulatePredicateRes, oldValue);
    }
    catch(e) {
      // just in case
      dump(e);
    }
  }

  // Get the domains sequence, or create it if it doesn't exist
  this.domainsContainer = this.rdfContainerUtils.MakeSeq(this.datasource, passwordMaker.domainsSubjectRes);

  this.assertGlobals();
  this.assertPost6Resources();
  this.datasourceFlusher.Flush();
}

// RDF member vars (containers are kept here; predicates and subjects are kept in passwdmaker.js)
PasswordMakerHelper.prototype.rdfContainer = Components.classes["@mozilla.org/rdf/container;1"].createInstance(Components.interfaces.nsIRDFContainer);
PasswordMakerHelper.prototype.rdfContainerUtils = Components.classes["@mozilla.org/rdf/container-utils;1"].getService(Components.interfaces.nsIRDFContainerUtils);
PasswordMakerHelper.prototype.rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
PasswordMakerHelper.prototype.datasource = null;
PasswordMakerHelper.prototype.accountsContainer = null;
PasswordMakerHelper.prototype.domainsContainer = null;
PasswordMakerHelper.prototype.datasourceFlusher = null;

// Other Member vars
PasswordMakerHelper.prototype.prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
PasswordMakerHelper.prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
PasswordMakerHelper._masterPassword = null;


PasswordMakerHelper.prototype.initDataSource = function() {
	var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
	// *** Get the resource from disk. Inspired by TBE and MenuEdit.
	var dir = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties).get("ProfD", Components.interfaces.nsILocalFile);
	var tempLocalFile = Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
	tempLocalFile.initWithPath(dir.path);
  var datasourceURI;

	try {
		datasourceURI = ioService.newFileURI(tempLocalFile).spec;
	}
	catch(e) {
    this.prompts.alert(window, "PasswordMaker", "Cannot create passwordmaker.rdf URI: " + e);
  }

	if (!datasourceURI.match(/\/$/))
    datasourceURI += "/";
  datasourceURI += "passwordmaker.rdf";
 // *** End TBE and MenuEdit inspiration.

  // Init the datasource, flusher, etc.
  this.datasource = this.rdfService.GetDataSourceBlocking(datasourceURI);
  this.datasourceFlusher = this.datasource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
}

PasswordMakerHelper.prototype.assertGlobals = function() {
  // Do the globalsettings exist? Make them if not. @TODO: change this code to use saveGlobals() instead
  if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordStoragePredicateRes))
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordStoragePredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("passwordStorage", "do-not-store")), true);
  if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.maskMasterPasswordPredicateRes))
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.maskMasterPasswordPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("maskGeneratedPassword", "false")), true);
  if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.autoClearClipboardPredicateRes))
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.autoClearClipboardPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("clearClipboard", "true")), true);
  if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.autoClearClipboardSecondsPredicateRes))
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.autoClearClipboardSecondsPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("clearClipboardTime", 10)), true);
  if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.advancedOptionsDialogPredicateRes))
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.advancedOptionsDialogPredicateRes, this.rdfService.GetLiteral(PasswordMakerPrefsWrapper.getChar("advancedOptionsDialog", "false")), true);
  if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.hideMasterPasswordFieldPredicateRes))
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.hideMasterPasswordFieldPredicateRes, this.rdfService.GetLiteral("false"), true); // 0.7.1
  if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.viewPwdFieldsPredicateRes))
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.viewPwdFieldsPredicateRes, this.rdfService.GetLiteral("false"), true); // 0.8.4
  if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.removePwdAutoCompletePredicateRes))
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.removePwdAutoCompletePredicateRes, this.rdfService.GetLiteral("false"), true); // 0.8.4
  if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.confirmMPWPredicateRes))
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.confirmMPWPredicateRes, this.rdfService.GetLiteral("true"), true); // 0.8.4

  // For people upgrading from versions where MPW was stored as a preference,
  // get the pref, store in the rdf, then clear the pref
  var mpw = PasswordMakerPrefsWrapper.getChar("masterPassword", null),
    mpwKey = PasswordMakerPrefsWrapper.getChar("key", null);
  if (mpw) {
    PasswordMakerPrefsWrapper.clear("masterPassword");
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordPredicateRes, this.rdfService.GetLiteral(mpw), true);
  }

  if (mpwKey) {
    PasswordMakerPrefsWrapper.clear("key");
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordKeyPredicateRes, this.rdfService.GetLiteral(mpwKey), true);
  }
}

/**
 * For backwards compatibility, ensure every account subject has a
 * charset predicate/target, suffix, and prefix
 * These were introduced in 0.6.1. This can be removed in a
 * future version when we're comfortable that all users are using
 * 0.6.1 or above.
 */
PasswordMakerHelper.prototype.assertPost6Resources = function() {
  var ret = new Array();
  var resEnum = this.datasource.GetAllResources();
  // Prior to 0.6.1, only the lower-case hex charset was available
  var hex = "0123456789abcdef";
  var hexLit = this.rdfService.GetLiteral(hex);
  var empty = this.rdfService.GetLiteral("");
  var falseLit = this.rdfService.GetLiteral("false");
  while (resEnum.hasMoreElements()) {
    var subjectRes = resEnum.getNext(); // e.g., http://passwordmaker.mozdev.org/yahoo/tom_cruise
    var whereLeetLB = this.datasource.GetTarget(subjectRes, passwordMaker.whereLeetLBPredicateRes, true);
    if (whereLeetLB) {
      // This is an account (not an account folder, globals, or anything else)

      /*    
       hasArcOut() isn't working as expected - duh, because i used passwordMaker.globalSettingsSubjectRes
       not subjectRes

      if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.charsetPredicateRes))
        this.datasource.Assert(subjectRes, passwordMaker.charsetPredicateRes, hexLit, true);
      if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.prefixPredicateRes))
        this.datasource.Assert(subjectRes, passwordMaker.prefixPredicateRes, empty, true);
      if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.suffixPredicateRes))
        this.datasource.Assert(subjectRes, passwordMaker.suffixPredicateRes, empty, true);*/
      
      var target = this.datasource.GetTarget(subjectRes, passwordMaker.charsetPredicateRes, true);
      if (!target)
        this.datasource.Assert(subjectRes, passwordMaker.charsetPredicateRes, hexLit, true);
      target = this.datasource.GetTarget(subjectRes, passwordMaker.prefixPredicateRes, true);
      if (!target)
        this.datasource.Assert(subjectRes, passwordMaker.prefixPredicateRes, empty, true);
      target = this.datasource.GetTarget(subjectRes, passwordMaker.suffixPredicateRes, true);
      if (!target)
        this.datasource.Assert(subjectRes, passwordMaker.suffixPredicateRes, empty, true);

      // 0.6.1 betas introduced a bug whereby multiple charset, prefix, and suffix targets could be written.
      // Check for them and fix if necessary.
      //this._fixMultipleTargets(subjectRes, passwordMaker.charsetPredicateRes, "character set", hex);
      //this._fixMultipleTargets(subjectRes, passwordMaker.prefixPredicateRes, "password prefix", "");
      //this._fixMultipleTargets(subjectRes, passwordMaker.suffixPredicateRes, "password suffix", "");      
    }
  }

  // Introduced 0.6.1
  if (!this.datasource.hasArcOut(passwordMaker.globalSettingsSubjectRes, passwordMaker.passwordShortcutPredicateRes))
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.passwordShortcutPredicateRes, this.rdfService.GetLiteral("pop-all"), true);
}

// Only called by assertPost6Resources() to fix a bug introduced by 0.6.1 beta
/*PasswordMakerHelper.prototype._fixMultipleTargets = function(subjectRes, predicateRes, action, newValue) {
  var targets = this.datasource.GetTargets(subjectRes, predicateRes, true);
  var acctName = this.datasource.GetTarget(subjectRes, passwordMaker.namePredicateRes, true); // get acct name
  if (acctName instanceof Components.interfaces.nsIRDFLiteral) {
    acctName = acctName.Value;
    var count = 0;
    while (targets.hasMoreElements()) {
      targets.getNext();
      count++;
    }
    if (count > 1) {
      count = 0;
      targets = this.datasource.GetTargets(subjectRes, predicateRes, true);
      var saved;
      while (targets.hasMoreElements()) {
        var target = targets.getNext();
        if (count++ == 0) {
          if (target instanceof Components.interfaces.nsIRDFLiteral)
            saved = target.Value;
        }
        else
          this.datasource.Unassert(subjectRes, predicateRes, target);
      }
      this.datasourceFlusher.Flush();
      if (saved.length > 40)
        saved = saved.substring(0, 40) + " [clipped for display]";
      this.prompts.alert(window, "PasswordMaker", "The account \"" + acctName + "\" has multiple " + action + "s defined but should only have one. This is the result of a bug introduced in PasswordMaker 0.6.1 beta. PasswordMaker repaired the problem by deleting the extra " + action + "s and using only\n\n" + saved + "\n\n as the " + action + " for this account. You should edit the settings for this account if that " + action + " is not what you want. We apologize for the inconvenience.");
    }
  }
}*/

/**
 * Make a string of 32 pseudo-random characters and return it.
 * This string is used as the key with which to encrypt the master
 * password (if the user decides to store it on the local machine).
 */
PasswordMakerHelper.prototype._makeKey32 = function() {
  var ret = "";
  while (ret.length < 32) {
    var rnd = Math.random().toString();
    ret += rnd.substring(rnd.lastIndexOf(".") + 1);
  }
  return ret.substring(0, 32);
}

PasswordMakerHelper.prototype.encrypt = function(clearText) {
  if (clearText == null || clearText == "")
    return null;
  key = PasswordMakerHelper.prototype._makeKey32();
  var encryptedData = byteArrayToHex(rijndaelEncrypt(clearText, hexToByteArray(key), "CBC"));
  return new MasterPassword(encryptedData, key);
}

PasswordMakerHelper.prototype.decrypt = function(encryptedMasterPassword, key) {
  //dump("encryptedMasterPassword="+encryptedMasterPassword+"\nkey="+key+"\n");
  if (!encryptedMasterPassword || !key)
    return null;

  var decryptedMasterPassword = byteArrayToString(rijndaelDecrypt(hexToByteArray(encryptedMasterPassword), hexToByteArray(key), "CBC"));
  //dump("Decrypted Master Password(1) = " + decryptedMasterPassword + "\n");
  return decryptedMasterPassword;
}

PasswordMakerHelper.prototype._decrypt = function(){
  return PasswordMakerHelper.prototype.decrypt(PasswordMakerHelper._masterPassword.encryptedData, PasswordMakerHelper._masterPassword.key);
}

/**
 * Generate the password from the key, a hash algorithm,
 * and (optionally) applying l33t speak afterwards. If we're
 * using an HMAC algorithm, we need a key and data, not just a key.
 */
PasswordMakerHelper.prototype.generatepassword = function(hashAlgorithm, key, data, whereToUseL33t, l33tLevel, passwordLength, charset, prefix, suffix) {
    
  // Never *ever, ever* allow the charset's length<2 else
  // the hash algorithms will run indefinitely
  if (charset == null || charset.length < 2 || hashAlgorithm.indexOf("0.6") > -1)
    charset = "0123456789abcdef";

  /*dump("Generating password for:\n");
  dump("   hashAlgorithm=" + hashAlgorithm + "\n");
  dump("   key=" + key + "\n");
  dump("   data=" + data + "\n");
  dump("   whereToUseL33t=" + whereToUseL33t + "\n");
  dump("   l33tLevel=" + l33tLevel + "\n");
  dump("   passwordLength=" + passwordLength + "\n");
  dump("   charset=" + charset + "\n");
  dump("   prefix=" + prefix + "\n");
  dump("   suffix=" + suffix + "\n");*/

  // for non-hmac algorithms, the key is master pw and url concatenated
  var usingHMAC = hashAlgorithm.indexOf("hmac") > -1;
  if (!usingHMAC)
    key += data; 

  // apply l33t before the algorithm?
  if (whereToUseL33t == "both" || whereToUseL33t == "before-hashing") {
    key = PasswordMaker_l33t.convert(l33tLevel, key);
    if (usingHMAC)
      data = PasswordMaker_l33t.convert(l33tLevel, data); // new for 0.3; 0.2 didn't apply l33t to _data_ for HMAC algorithms
  }

  // apply the algorithm
  var password;
  switch(hashAlgorithm) {
    case "sha256":
      password = PasswordMaker_SHA256.any_sha256(key, charset);
      break;
    case "hmac-sha256":
      password = PasswordMaker_SHA256.any_hmac_sha256(key, data, charset);
      break;
    case "sha1":
      password = PasswordMaker_SHA1.any_sha1(key, charset);
      break;
    case "hmac-sha1":
      password = PasswordMaker_SHA1.any_hmac_sha1(key, data, charset);
      break;
    case "md4":
      password = PasswordMaker_MD4.any_md4(key, charset);
      break;
    case "hmac-md4":
      password = PasswordMaker_MD4.any_hmac_md4(key, data, charset);
      break;
    case "md5":
      password = PasswordMaker_MD5.any_md5(key, charset);
      break;
    case "md5-v0.6":
      password = PasswordMaker_MD5_V6.hex_md5(key);
      break;      
    case "hmac-md5":
      password = PasswordMaker_MD5.any_hmac_md5(key, data, charset);
      break;
    case "hmac-md5-v0.6":
      password = PasswordMaker_MD5_V6.hex_hmac_md5(key, data);
      break;       
    case "rmd160":
      password = PasswordMaker_RIPEMD160.any_rmd160(key, charset);
      break;
    case "hmac-rmd160":
      password = PasswordMaker_RIPEMD160.any_hmac_rmd160(key, data, charset);
      break;
    default:
      password = "coding error; contact author";
      return;
  }

  // apply l33t after the algorithm?
  if (whereToUseL33t == "both" || whereToUseL33t == "after-hashing")
    password = PasswordMaker_l33t.convert(l33tLevel, password);
  if (prefix)
    password = prefix + password;
  if (suffix) {
    var pw = password.substring(0, passwordLength-suffix.length) + suffix;
    return pw.substring(0, passwordLength);
  }
  //dump("   generated password=" + password.substring(0, passwordLength) + "\n\n");
  return password.substring(0, passwordLength);
}

/**
 * Function for displaying dialog box with yes/no buttons (not OK/Cancel buttons).
 */
PasswordMakerHelper.prototype.ask = function(parent, text) {
  return PasswordMakerHelper.prompts.confirmEx(parent, "PasswordMaker", text,
                                PasswordMakerHelper.prompts.BUTTON_TITLE_IS_STRING * PasswordMakerHelper.prompts.BUTTON_POS_0 +
                                PasswordMakerHelper.prompts.BUTTON_TITLE_IS_STRING * PasswordMakerHelper.prompts.BUTTON_POS_1,
                                "Yes", "No", null, null, {}) == 0; // 0 means first button ("yes") was pressed
}

/**
 * Returns the decrypted master password string or null.
 * Prompt the user if we don't have a _masterPassword object.
 */
PasswordMakerHelper.prototype.getDecryptedMasterPassword = function() {
  if (PasswordMakerHelper._masterPassword)
    return PasswordMakerHelper.prototype._decrypt();
  else
    return PasswordMakerHelper.prototype._passwordPrompt();
}

/**
 * Set the _masterPassword variable from the supplied argument.
 * This supplied argument must be a MasterPassword object.
 */
PasswordMakerHelper.prototype.setEncryptedMasterPassword = function(oMasterPassword) {
  PasswordMakerHelper._masterPassword = oMasterPassword && oMasterPassword.encryptedData && oMasterPassword.key ? oMasterPassword : null;
  //dump(PasswordMakerHelper._masterPassword.encryptedData + "\n");
  //dump(PasswordMakerHelper._masterPassword.key + "\n");
}

/**
 * Return true if a master password has been set,
 * false if not.
 */
PasswordMakerHelper.prototype.hasMasterPassword = function() {
  //dump("_masterPassword is " + PasswordMakerHelper._masterPassword + "\n");
  return PasswordMakerHelper._masterPassword != null;
}

PasswordMakerHelper.prototype.storeOrClearMasterPassword = function(action, clearTextMasterPassword) {
  var helper = passwordMaker.passwordMakerHelper;
  var mpwRes = helper.datasource.GetTarget(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordPredicateRes, true);
  var mpwKeyRes = helper.datasource.GetTarget(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordKeyPredicateRes, true);

  if (action == "do-not-store" || clearTextMasterPassword == "" || clearTextMasterPassword == null) {
    // Clear any existing master pw (encrypted form) and its key, if one exists.
    // From disk...
    if (mpwRes != null)
      helper.datasource.Unassert(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordPredicateRes, mpwRes);
    if (mpwKeyRes != null)
      helper.datasource.Unassert(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordKeyPredicateRes, mpwKeyRes);

    // From memory...
    PasswordMakerHelper.prototype.setEncryptedMasterPassword(null);
  }
  else {
    // Store the master pw in memory (encrypted form)
    var obj = PasswordMakerHelper.prototype.encrypt(clearTextMasterPassword);
    PasswordMakerHelper.prototype.setEncryptedMasterPassword(obj); // update master pw/key 

    if (action == "store-on-disk" && obj != null) {
      // Store the master pw (encypted form) on disk
      if (mpwRes == null)
        helper.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordPredicateRes, helper.rdfService.GetLiteral(obj.encryptedData), true);
      else
        helper.datasource.Change(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordPredicateRes, mpwRes, helper.rdfService.GetLiteral(obj.encryptedData));

      if (mpwKeyRes == null)
        helper.datasource.Assert(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordKeyPredicateRes, helper.rdfService.GetLiteral(obj.key), true);
      else
        helper.datasource.Change(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordKeyPredicateRes, mpwKeyRes, helper.rdfService.GetLiteral(obj.key));
    }
  }
}

/**
 * Prompt user for the master password, ask if he wants to save it, etc.
 * Return the master pw in clear text or null if user cancelled.
 * Also, store the master password in the PasswordMakerHelper._masterPassword variable
 * if the user doesn't want to be prompted anymore.
 */
PasswordMakerHelper.prototype._passwordPrompt = function() {

  var params = {inn:{passwordStorageLBValue :
                     passwordMaker._getLiteral(passwordMaker.globalSettingsSubjectRes, passwordMaker.masterPasswordStoragePredicateRes),
                     passwordMakerHelper : passwordMaker.passwordMakerHelper, passwordMaker : passwordMaker},
                out:null};

  // Return null or a decrypted password
    // User clicked OK
  while (true) {
    window.openDialog("chrome://passwdmaker/chrome/passwordprompt.xul",
     "", "chrome, dialog, modal, resizable=yes", params).focus();
    if (params.out == null) {
      // user canceled
      return null;
    }
    else if (params.out.masterPassword == "") {
      // mpw can't be empty
      passwordMaker.passwordMakerHelper.prompts.alert(this, "PasswordMaker", "The master password cannot be empty.");
      params.out = null;
    }
    else {
      //PasswordMakerHelper.prototype.saveGlobalSetting("masterPasswordStorage", params.out.passwordStorage);    
      PasswordMakerHelper.prototype.storeOrClearMasterPassword(params.out.passwordStorage, params.out.masterPassword);
      return params.out.masterPassword;
    }
  }
}

/**
 * Save a global setting using Assert() or Change() appropriately.
 */
PasswordMakerHelper.prototype.saveGlobalSetting = function(predicateStr, newLiteralStr) {
  var newTargetRes = this.rdfService.GetLiteral(newLiteralStr);
  var predicateRes = this.rdfService.GetResource("http://passwordmaker.mozdev.org/rdf#" + predicateStr);
  var oldTargetRes;
  try {
    oldTargetRes = this.datasource.GetTarget(passwordMaker.globalSettingsSubjectRes, predicateRes, true);
  }
  catch (e) {
    // Old target doesn't exist. Do an Assert() instead of Change()
    this.datasource.Assert(passwordMaker.globalSettingsSubjectRes, predicateRes, newTargetRes, true);
    this.datasourceFlusher.Flush();
    return;      
  }
  this.datasource.Change(passwordMaker.globalSettingsSubjectRes, predicateRes, oldTargetRes, newTargetRes);
  this.datasourceFlusher.Flush();
}

/**
 * Launch the native application associated with mailto: links.
 */
PasswordMakerHelper.prototype.sendMail = function(mailtoUrl) {
  // Copied from browser.js's MailIntegration class

  var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
  var uri = ioService.newURI(mailtoUrl, null, null);

  // Now pass this url to the operating system
  // a generic method which can be used to pass arbitrary urls to the operating system.
  // aURL --> a nsIURI which represents the url to launch
  var extProtocolSvc = Components.classes["@mozilla.org/uriloader/external-protocol-service;1"].getService(Components.interfaces.nsIExternalProtocolService);
  if (extProtocolSvc)
      extProtocolSvc.loadUrl(uri);
}

// Prompt user for the master password before.
// Return t/f.
PasswordMakerHelper.prototype.prompwMPW = function(mpw) {
  var okclicked = true;
  try {
    while (okclicked) {
      var input = {value:""};
      okclicked = this.prompts.promptPassword(this, "PasswordMaker",
        "For security purposes, you must enter the master password:", input, null, {});
      if (okclicked) {
        if (input.value == mpw)
          return true;
        else
          this.prompts.alert(this, "PasswordMaker", "That is not the correct master password.");
      }
    }
    return false;
  }
  finally {
    mpw = null; // proactively clear sensitive memory
  }
}